# !/usr/bin/python3

import os
import sys
import time

import multiprocessing
import tempfile
import numpy as np

from pdb_eda import ccp4

ccp4urlPrefix = "http://www.ebi.ac.uk/pdbe/coordinates/files/"
ccp4urlSuffix = ".ccp4"
ccp4folder = './ccp4_data/'

def fromPDBid(pdbid):
    """
    Calculate background density
    """
    pdbid = pdbid.lower()
    ccp4diffFile = ccp4folder + pdbid + '_diff.ccp4'
    
    if os.path.isfile(ccp4diffFile):
        diffDensityObj = ccp4.read(ccp4diffFile, pdbid)
        if diffDensityObj is None or diffDensityObj == 0:
            return None
        
        diffDensityCutoff = np.mean(diffDensityObj.densityArray) + 3 * np.std(diffDensityObj.densityArray)
        totalDensity = np.sum([abs(i) for i in diffDensityObj.densityArray if i > diffDensityCutoff or i < -1 * diffDensityCutoff])
        volume = diffDensityObj.header.unitVolume * diffDensityObj.header.nintervalX * diffDensityObj.header.nintervalY * diffDensityObj.header.nintervalZ
        sphereV = 179.5943800302165 # 4 * np.pi * 3.5 ** 3 / 3
        bkgdDensity = totalDensity / volume * sphereV
        
        return [pdbid, totalDensity, volume, bkgdDensity]
    else:
        return None

def processFunction(pdbid):
    result = fromPDBid(pdbid)

    if result:
        item = ','.join([str(j) for j in result])
        globalTempFile.write("%s\n" % item)
    return globalTempFile.name

def openTemporaryFile(temp):
    dirname = os.getcwd()
    global globalTempFile
    globalTempFile = tempfile.NamedTemporaryFile(mode='w', buffering=1, dir=dirname, prefix="tempBG_", delete=False)
    time.sleep(0.01) # sleep 0.01 seconds to prevent the same worker from calling this twice.
    return globalTempFile.name

def closeTemporaryFile(temp):
    filename = globalTempFile.name
    globalTempFile.close()
    # Sleep 0.01 seconds to prevent the same worker from calling this twice.
    # May need to increase this to 1 second or even 10 seconds to make this work.
    time.sleep(0.01)
    return filename

if __name__ == "__main__":
    pdbidFile = sys.argv[1]
    resultFile = sys.argv[2]

    pdbids = []
    with open(pdbidFile, "r") as fileHandleIn:
        for pdbid in fileHandleIn:
            pdbids.append(pdbid[0:4])

    with multiprocessing.Pool() as pool:
        result_filenames = pool.map(openTemporaryFile, (1 for i in range(multiprocessing.cpu_count())))
        pool.map(processFunction, pdbids)
        closed_filenames = pool.map(closeTemporaryFile, (1 for i in range(multiprocessing.cpu_count())))

    unclosed_filenames = set(result_filenames) - set(closed_filenames)
    if unclosed_filenames: # check whether all result files were closed.  Increase sleep time in closeTemporary File to prevent this.
        print("Unclosed Files: ", unclosed_filenames)

    with open(resultFile, "w") as outfile:
        for f in result_filenames:
            if f:
                with open(f, "r") as infile:
                    outfile.write(infile.read())
                os.remove(f) # remove the file

