# !/usr/bin/python3

import os
import sys
import time

import multiprocessing
import tempfile
import numpy as np
import Bio.PDB as biopdb
from pdb_eda import pdbParser

clusterCutoff = 3.5
pdbfolder = './pdb_data/'
metalList = ['ZN', 'MG', 'CA', 'FE', 'NA', 'MN', 'K', 'CU', 'NI', 'CO', 'CD', 'HG', 'PT', 'MO', 'AL', 'V',
             'BA', 'SR', 'RU', 'CS', 'W', 'YB', 'AU', 'Y', 'LI', 'PB', 'GD', 'TL', 'RB', 'SM', 'IR', 'PR',
             'RH', 'EU', 'PD', 'AG', 'OS', 'LU', 'HO', 'TB', 'CR', 'GA', 'LA', 'SB', 'CE', 'ER', 'IN', 'BI',
             'DY', 'BE', 'ZR', 'SN', 'HF', 'TA', 'RE', 'PA', 'U']

def fromPDBid(pdbid):
    """
    Calculate background density
    """
    pdbid = pdbid.lower()
    pdbfile = pdbfolder + 'pdb' + pdbid + '.ent'
    
    results = []
    if os.path.isfile(pdbfile):
        parser = biopdb.PDBParser(QUIET=True)
        biopdbObj = parser.get_structure(pdbid, pdbfile)
        
        if not biopdbObj:
            return None
        else:
            metals = [i for i in biopdbObj.get_atoms() if i.element in metalList]
            if len(metals) <= 1:
                return None
            elif len(metals) == 2:
                dist = np.linalg.norm(metals[0].coord - metals[1].coord)
                results.append([pdbid, metals[0].parent.parent.id, metals[0].parent.id[1], metals[0].name, metals[0].element, dist])
                results.append([pdbid, metals[1].parent.parent.id, metals[1].parent.id[1], metals[1].name, metals[1].element, dist])
            else:
                for metal in metals:
                    try:
                        minDist = np.min([np.linalg.norm(atom.coord - metal.coord) for atom in metals if not np.array_equal(atom.coord, metal.coord)])
                        results.append([pdbid, metal.parent.parent.id, metal.parent.id[1], metal.name, metal.element, minDist])
                    except ValueError:
                        print(pdbid, metal.parent.parent.id, metal.parent.id[1], metal.name, metal.element, metal.coord)
            return results
    else:
        return None

def processFunction(pdbid):
    results = fromPDBid(pdbid)
    if results:
        results = [','.join([str(j) for j in i]) for i in results]

        for item in results:
            globalTempFile.write("%s\n" % item)

    return globalTempFile.name

def openTemporaryFile(temp):
    dirname = os.getcwd()
    global globalTempFile
    globalTempFile = tempfile.NamedTemporaryFile(mode='w', buffering=1, dir=dirname, prefix="tempMC_", delete=False)
    time.sleep(0.01) # sleep 0.01 seconds to prevent the same worker from calling this twice.
    return globalTempFile.name

def closeTemporaryFile(temp):
    filename = globalTempFile.name
    globalTempFile.close()
    # Sleep 0.01 seconds to prevent the same worker from calling this twice.
    # May need to increase this to 1 second or even 10 seconds to make this work.
    time.sleep(0.01)
    return filename

if __name__ == "__main__":
    pdbidFile = sys.argv[1]
    resultFile = sys.argv[2]

    pdbids = []
    with open(pdbidFile, "r") as fileHandleIn:
        for pdbid in fileHandleIn:
            pdbids.append(pdbid[0:4])

    with multiprocessing.Pool() as pool:
        result_filenames = pool.map(openTemporaryFile, (1 for i in range(multiprocessing.cpu_count())))
        pool.map(processFunction, pdbids)
        closed_filenames = pool.map(closeTemporaryFile, (1 for i in range(multiprocessing.cpu_count())))

    unclosed_filenames = set(result_filenames) - set(closed_filenames)
    if unclosed_filenames: # check whether all result files were closed.  Increase sleep time in closeTemporary File to prevent this.
        print("Unclosed Files: ", unclosed_filenames)

    with open(resultFile, "w") as outfile:
        for f in result_filenames:
            if f:
                with open(f, "r") as infile:
                    outfile.write(infile.read())
                os.remove(f) # remove the file

