# !/usr/bin/python3

"""
Application to create a high-quality metalloprotein sites (pdb_eda.metalloprotein)
-------------------------------------------------------

This module provides methods to find and filter out low-quality metal binding sites from PDB id,
"""
import sys
import os
import sys
import time

import numpy as np
import multiprocessing
import tempfile
import Bio.PDB as biopdb
from pdb_eda import pdbParser
from pdb_eda import ccp4

metalRadius = 3.5
ccp4urlPrefix = "http://www.ebi.ac.uk/pdbe/coordinates/files/"
ccp4urlSuffix = ".ccp4"
ccp4folder = './ccp4_data/'
pdbfolder = './pdb_data/'
metalList = ['ZN', 'MG', 'CA', 'FE', 'NA', 'MN', 'K', 'CU', 'NI', 'CO', 'CD', 'HG', 'PT', 'MO', 'AL', 'V',
          'BA', 'SR', 'RU', 'CS', 'W', 'YB', 'AU', 'Y', 'LI', 'PB', 'GD', 'TL', 'RB', 'SM', 'IR', 'PR',
          'RH', 'EU', 'PD', 'AG', 'OS', 'LU', 'HO', 'TB', 'CR', 'GA', 'LA', 'SB', 'CE', 'ER', 'IN', 'BI', 
          'DY', 'BE', 'ZR', 'SN', 'HF', 'TA', 'RE', 'PA', 'U']

def fromPDBid(pdbid):
    """
    Creates :class:`pdb_eda.metalloprotein` object given the PDB id if the id is valid.

    :param str pdbid: PDB id
    :return: :class:`pdb_eda.metalloprotein`
    """
    pdbid = pdbid.lower()
    pdbfile = pdbfolder + 'pdb' + pdbid + '.ent'
    ccp4diffFile = ccp4folder + pdbid + '_diff.ccp4'
    results = []
    resolution = occupancy = symmetry = density = 0
    diffDensityObj = 0

   
    if not os.path.isfile(pdbfile):  # No pdb structure file
        return [[pdbid, -1, 0, 0, 0, 0, 0, 0, 0]]
    else:
        parser = biopdb.PDBParser(QUIET=True)
        biopdbObj = parser.get_structure(pdbid, pdbfile)
        pdbObj = pdbParser.readPDBfile(pdbfile, 'full')

        if not biopdbObj:
            return [[pdbid, -2, 0, 0, 0, 0, 0, 0, 0]]  # Error reading pdb file
        elif biopdbObj.header['resolution'] is None: 
            resolution = -3  # No resolution information, e.g. NMR resolved structures
        else:
            resolution = biopdbObj.header['resolution']

        metals = [i for i in biopdbObj.get_atoms() if i.element in metalList]
        if len(metals) == 0:  # No metal binding sites
            results.append([pdbid, resolution, 0, 0, 0, 0, 0, 0, 0])
        else:
            symmetryAtoms = pdbObj.calcSymmetryAtoms([i.coord for i in metals])

            if os.path.isfile(ccp4diffFile):
                diffDensityObj = ccp4.read(ccp4diffFile, pdbid)

            for metal in metals:
                if len(symmetryAtoms) == 0:
                    symmetry = -1  # No symmetry atoms data
                elif np.min([np.linalg.norm(atom.coord - metal.coord) for atom in symmetryAtoms]) <= metalRadius:
                    symmetry = 1  # Symmetry atoms nearby
                else:
                    symmetry = 0  # No symmetry atoms nearby

                if diffDensityObj is None:
                    density = -2  # Error in reading ccp4 file 
                elif diffDensityObj == 0:
                    density = -1  # No ccp4 files
                else:
                    diffDensityObj.diffDensityCutoff = np.mean(diffDensityObj.densityArray) + 3 * np.std(diffDensityObj.densityArray)
                    green = diffDensityObj.findAberrantBlobs(metal.get_coord(), metalRadius, diffDensityObj.diffDensityCutoff)
                    red = diffDensityObj.findAberrantBlobs(metal.get_coord(), metalRadius, -1 * diffDensityObj.diffDensityCutoff)

                    density = sum([abs(i.totalDensity) for i in green + red]) # Total density around the metal

                results.append([pdbid, resolution, metal.parent.parent.id, metal.parent.id[1], metal.name, metal.element, metal.get_occupancy(), symmetry, density])

    if len(results) == 0:
        print(pdbid)
        results.append([pdbid, -2, 0, 0, 0, 0, 0, 0, 0])

    return results

def processFunction(pdbid):
    results = fromPDBid(pdbid)
    results = [','.join([str(j) for j in i]) for i in results]
    
    for item in results:
        globalTempFile.write("%s\n" % item)
    return globalTempFile.name

def openTemporaryFile(temp):
    dirname = os.getcwd()
    global globalTempFile
    globalTempFile = tempfile.NamedTemporaryFile(mode='w', buffering=1, dir=dirname, prefix="tempMP_", delete=False)
    time.sleep(0.01) # sleep 0.01 seconds to prevent the same worker from calling this twice.
    return globalTempFile.name

def closeTemporaryFile(temp):
    filename = globalTempFile.name
    globalTempFile.close()
    # Sleep 0.01 seconds to prevent the same worker from calling this twice.
    # May need to increase this to 1 second or even 10 seconds to make this work.
    time.sleep(0.01)
    return filename

if __name__ == "__main__":
    pdbidFile = sys.argv[1]
    resultFile = sys.argv[2]
    
    pdbids = []
    with open(pdbidFile, "r") as fileHandleIn:
        for pdbid in fileHandleIn:
            pdbids.append(pdbid[0:4])

with multiprocessing.Pool() as pool:
    result_filenames = pool.map(openTemporaryFile, (1 for i in range(multiprocessing.cpu_count())))
    pool.map(processFunction, pdbids)
    closed_filenames = pool.map(closeTemporaryFile, (1 for i in range(multiprocessing.cpu_count())))
    
    unclosed_filenames = set(result_filenames) - set(closed_filenames)
    if unclosed_filenames: # check whether all result files were closed.  Increase sleep time in closeTemporary File to prevent this.
        print("Unclosed Files: ", unclosed_filenames)
    
    with open(resultFile, "w") as outfile:
        for f in result_filenames:
            if f:
                with open(f, "r") as infile:
                    outfile.write(infile.read())
                os.remove(f) # remove the file


